import eric_sse
from eric_sse.message import MessageContract

logger = eric_sse.get_logger()

class MessageQueueListener:
    """
    Base class for listeners.

    Optionally you can override on_message method if you need to inject code at message delivery time.
    """

    def __init__(self):
        self.id: str = eric_sse.generate_uuid()
        self.__is_running: bool = False

    def on_message(self, msg: MessageContract) -> None:
        """Event handler. It executes when a message is delivered to client"""
        pass

    def start(self) -> None:
        self.__is_running = True

    def stop(self) -> None:
        self.__is_running = False

    def is_running(self) -> bool:
        return self.__is_running




"""
This module is intended to those who want to create their own persistence layer

A **Redis** concrete implementation of interfaces is available at  `eric-redis-queues package source <https://github.com/laxertu/eric-redis-queues/blob/master/eric_redis_queues/__init__.py>`_.

Here is a `Basic example <https://github.com/laxertu/eric/blob/master/examples/use_cases/basic.py>`_ of integration.

**Writing a custom persistence layer**

You'll need to implement the following interfaces:

**Channels**

* :class:`~eric_sse.persistence.ChannelRepositoryInterface`

* You'll need to define a channel that implements :class:`~eric_sse.persistence.PersistableChannel` if :class:`~eric_sse.prefabs.SSEChannel` do not suit with your requirements
* For **MessageQueueListener** support you can extend or directly use :class:`~eric_sse.persistence.PersistableListener`.

**Connections**

* :class:`~eric_sse.persistence.PersistableQueue`
* :class:`~eric_sse.persistence.ConnectionRepositoryInterface`
"""

from abc import ABC, abstractmethod
from typing import Iterable, Any
from importlib import import_module

from eric_sse.connection import Connection
from eric_sse.listener import MessageQueueListener
from eric_sse.queues import Queue, InMemoryQueue


class ObjectAsKeyValuePersistenceMixin(ABC):
    """
    Adds KV persistence support.

    By implementing this abstract mixin should be possible to persist every object that is not directly
    serializable by pickle, for example, if your Queues implementation wraps some incompatible dependency, e.g. a Redis client.

    For this reason, the idea is that dict values should be serializable by pickle too.

    see :func:`~eric_sse.persistence.importlib_create_instance`
    """
    @property
    @abstractmethod
    def kv_key(self) -> str:
        """The key to use when persisting object"""
        ...

    @property
    @abstractmethod
    def kv_value_as_dict(self) -> dict:
        """Returns value that will be persisted as a dictionary."""
        ...

    @abstractmethod
    def setup_by_dict(self, setup: dict):
        """Does necessary post-creation setup of object given its persisted values"""
        ...

    @property
    def kv_class_absolute_path(self) -> str:
        """Returns class full path as string"""
        return f'{self.__module__}.{type(self).__name__}'

    @property
    @abstractmethod
    def kv_constructor_params_as_dict(self) -> dict:
        """Class constructor parameters as dict"""
        ...


def importlib_create_instance(class_full_path: str, constructor_params: dict, setup_values: dict) -> ObjectAsKeyValuePersistenceMixin:
    """
    Creates a persistable class instance given a persisted value and executes its setup_by_dict method

    see :func:`~eric_sse.persistence.ObjectAsKeyValuePersistenceMixin.setup_by_dict`
    """
    path_parts = class_full_path.split('.')
    module = '.'.join(path_parts[:-1])
    klass = path_parts[-1]

    module_object = import_module(module)
    obj: ObjectAsKeyValuePersistenceMixin = getattr(module_object, klass)(constructor_params)
    obj.setup_by_dict(setup_values)

    return obj


class PersistableQueue(Queue, ObjectAsKeyValuePersistenceMixin, ABC):
    """Concrete implementations of methods should perform in **Queues** ones their I/O operations, and define in **ObjectAsKeyValuePersistenceMixin** ones their correspondant persistence strategy"""
    ...

class PersistableListener(MessageQueueListener, ObjectAsKeyValuePersistenceMixin):
    """Gives KV persistence support to MessageQueueListener."""
    @property
    def kv_key(self) -> str:
        return self.id

    @property
    def kv_value_as_dict(self) -> dict:
        return {}

    def setup_by_dict(self, setup: dict):
        pass

    @property
    def kv_constructor_params_as_dict(self) -> dict:
        return {}


class PersistableConnection(Connection):
    listener: PersistableListener
    queue: PersistableQueue

class PersistableChannel(ObjectAsKeyValuePersistenceMixin, ABC):
    ...

class ObjectRepositoryInterface(ABC):
    """Every exception raised by concrete implementations show be wrapped inside a :class:`~eric_sse.exception.RepositoryError`
    """

    @abstractmethod
    def load(self) -> Iterable[ObjectAsKeyValuePersistenceMixin]:
        """Returns an Iterable of all persisted objects of correspondant concrete implementation."""
        ...

    @abstractmethod
    def persist(self, persistable: ObjectAsKeyValuePersistenceMixin):
        ...

    @abstractmethod
    def delete(self, key: str):
        ...


class ChannelRepositoryInterface(ObjectRepositoryInterface):

    @abstractmethod
    def delete_listener(self, channel_id: str, listener_id: str) -> None:
        ...

class ConnectionRepositoryInterface(ABC):
    """
    Abstraction for connections creation.

    It exposes methods to be used by ChannelRepositoryInterface implementations for connections loading.

    see :class:`~eric_sse.entities.AbstractChannel`
    """

    @abstractmethod
    def create_queue(self, listener_id: str) -> Queue:
        """
        Returns a concrete Queue instance.

        :param str listener_id: Corresponding listener id
        """
        ...

    @abstractmethod
    def persist(self, channel_id: str, connection: PersistableConnection) -> None:
        ...

    @abstractmethod
    def load_all(self) -> Iterable[PersistableConnection]:
        """Returns an Iterable of all persisted connections"""
        ...

    @abstractmethod
    def load(self, channel_id: str) -> Iterable[PersistableConnection]:
        """Returns an Iterable of all persisted connections of a given channel"""
        ...

    @abstractmethod
    def delete(self, channel_id: str, listener_id: str) -> None:
        """Removes a persisted :class:`~eric_sse.persistence.PersistableConnection` given its correspondant listener id"""
        ...



class InMemoryConnectionRepository(ConnectionRepositoryInterface):
    """
    Default implementation used by :class:`~eric_sse.entities.AbstractChannel`
    """
    def create_queue(self, listener_id: str) -> Queue:
        return InMemoryQueue()

    def persist(self, channel_id: str, connection: Connection) -> None:
        pass

    def load_all(self) ->  Iterable[Connection]:
        return []

    def load(self, channel_id: str) -> Iterable[Connection]:
        return []

    def delete(self, channel_id: str, listener_id: str) -> None:
        pass



class InvalidChannelException(Exception):
    ...


class InvalidListenerException(Exception):
    ...


class NoMessagesException(Exception):
    """Raised when trying to fetch from an empty queue"""
    ...


class InvalidMessageFormat(Exception):
    ...


class RepositoryError(Exception):
    """
    Raised when an unexpected error occurs while trying to fetch messages from a queue.

    Concrete implementations of :class:`~eric_sse.persistence.ObjectRepositoryInterface` should wrap here the unexpected exceptions they catch before raising.
    """
    ...



import asyncio
import traceback
from abc import ABC, abstractmethod
from typing import AsyncIterable, Any

import eric_sse
from eric_sse.exception import InvalidListenerException, NoMessagesException, InvalidChannelException
from eric_sse.listener import MessageQueueListener
from eric_sse.message import MessageContract, Message
from eric_sse.queues import Queue
from eric_sse.persistence import (ConnectionRepositoryInterface, InMemoryConnectionRepository,
                                  PersistableListener, PersistableConnection)

logger = eric_sse.get_logger()

MESSAGE_TYPE_CLOSED = '_eric_channel_closed'
MESSAGE_TYPE_END_OF_STREAM = '_eric_channel_eof'
MESSAGE_TYPE_INTERNAL_ERROR = '_eric_error'


class _ConnectionManager:
    """Maintains relationships between listeners and queues"""
    def __init__(self, channel_id: str, queues_repository: ConnectionRepositoryInterface):
        self.__channel_id = channel_id
        self.__listeners: dict[str: MessageQueueListener] = {}
        self.__queues: dict[str: Queue] = {}
        self.__queues_repository = queues_repository


    def load(self):
        for c in self.__queues_repository.load(self.__channel_id):
            self.__listeners[c.listener.id] = c.listener
            self.__queues[c.listener.id] = c.queue

    def add_listener(self) -> PersistableListener:
        l = PersistableListener()
        self.register_listener(l)
        return l

    def register_listener(self, listener: PersistableListener):

        queue = self.__queues_repository.create_queue(listener.id)

        self.__listeners[listener.id] = listener
        self.__queues[listener.id] = queue
        self.__queues_repository.persist(self.__channel_id, PersistableConnection(listener=listener, queue=queue))

    def register_connection(self, listener: MessageQueueListener, queue: Queue):
        self.__listeners[listener.id] = listener
        self.__queues[listener.id] = queue


    def remove_listener(self, listener_id: str):
        del self.__queues[listener_id]
        del self.__listeners[listener_id]

        self.__queues_repository.delete(channel_id=self.__channel_id, listener_id=listener_id)

    def get_queue(self, listener_id: str) -> Queue:
        try:
            return self.__queues[listener_id]
        except KeyError:
            raise InvalidListenerException(f"Invalid listener {listener_id}") from None

    def get_listener(self, listener_id: str) -> MessageQueueListener:
        try:
            return self.__listeners[listener_id]
        except KeyError:
            raise InvalidListenerException

    def get_listeners(self) -> dict[str, MessageQueueListener]:
        """Returns a dict mapping listener ids to listeners"""
        return self.__listeners

class AbstractChannel(ABC):
    """
    Base class for channels.

    Provides functionalities for listeners and message delivery management.

    **Important** When using persistence layer you have to call to **load_persisted_data()** method just after object creation.

    :class:`~eric_sse.persistence.InMemoryConnectionRepository` is the default implementation used for **connections_repository** parameter.

    see :class:`~eric_sse.prefabs.SSEChannel`

    :param int stream_delay_seconds: Wait time in seconds between message delivery.
    :param ~eric_sse.persistence.ConnectionRepositoryInterface connections_repository:
    :param str channel_id: Optionally sets the channel id. **IMPORTANT** by using this parameter, client is responsible for guaranteeing channel id uniqueness
    """
    def __init__(
            self,
            stream_delay_seconds: int = 0,
            connections_repository: ConnectionRepositoryInterface | None = None,
            channel_id: str | None = None
    ):
        self.__id: str = eric_sse.generate_uuid() if channel_id is None else channel_id
        assert self.__id is not None
        self.stream_delay_seconds = stream_delay_seconds

        self._connections_repository = connections_repository = connections_repository if connections_repository else InMemoryConnectionRepository()
        self.__connection_manager: _ConnectionManager = _ConnectionManager(self.__id, connections_repository)

    def load_persisted_data(self):
        """
        Loads persisted Connections

        see :class:`~eric_sse.persistence.ConnectionRepositoryInterface`
        """
        self.__connection_manager.load()

    @property
    def id(self) -> str:
        """Unique identifier for this channel, it can be set by **channel_id** constructor parameter"""
        return self.__id


    @abstractmethod
    def adapt(self, msg: MessageContract) -> Any:
        """Models output of channel streams"""
        ...

    async def message_stream(self, listener: MessageQueueListener) -> AsyncIterable[Any]:
        """
        Entry point for message streaming

        A message with type = 'error' is yield on invalid listener
        """

        async def new_messages():
            try:
                result = self.deliver_next(listener.id)
                yield result
            except NoMessagesException:
                ...

        async def event_generator() -> AsyncIterable[dict]:

            while True:
                # If client closes connection, stop sending events
                if not listener.is_running():
                    logger.debug("Listener stopped. Exiting")
                    break

                try:
                    async for message in new_messages():
                        yield self.adapt(message)

                    await asyncio.sleep(self.stream_delay_seconds)
                except (InvalidListenerException, InvalidChannelException) as e:
                    yield self.adapt(Message(msg_type='error', msg_payload=e))
                except Exception as e:
                    logger.debug(traceback.format_exc())
                    logger.error(e)
                    yield self.adapt(Message(msg_type='error'))

        async for event in event_generator():
            yield event

    def add_listener(self) -> MessageQueueListener:
        """Add the default listener and creates corresponding queue"""
        l = PersistableListener()
        self.register_listener(l)
        return l

    def register_listener(self, listener: PersistableListener):
        """
        Registers listener and creates corresponding queue with persistence support
        """
        return self.__connection_manager.register_listener(listener)

    def register_connection(self, listener: MessageQueueListener, queue: Queue):
        """Registers a Connection with listener and queue without persistence"""
        return self.__connection_manager.register_connection(listener, queue)

    def remove_listener(self, listener_id: str):
        self.__connection_manager.remove_listener(listener_id)

    def deliver_next(self, listener_id: str) -> MessageContract:
        """
        Returns next message for given listener id.

        Raises a NoMessagesException if queue is empty
        """
        listener = self.get_listener(listener_id)
        if listener.is_running():
            queue = self.__connection_manager.get_queue(listener.id)
            msg = queue.pop()
            listener.on_message(msg)
            return msg

        raise NoMessagesException

    def _get_queue(self, listener_id: str) -> Queue:
        return self.__connection_manager.get_queue(listener_id)

    def dispatch(self, listener_id: str, msg: MessageContract):
        """Adds a message to listener's queue"""

        queue = self._get_queue(listener_id)
        queue.push(msg)
        logger.debug(f"Dispatched {msg} to {listener_id}")

    def broadcast(self, msg: MessageContract):
        """Enqueue a message to all listeners"""
        for listener_id in self.__connection_manager.get_listeners():
            self.dispatch(listener_id, msg=msg)

    def get_listener(self, listener_id: str) -> MessageQueueListener:
        return self.__connection_manager.get_listener(listener_id)

    async def watch(self) -> AsyncIterable[Any]:
        # TODO tests
        listener = self.add_listener()
        listener.start()
        return self.message_stream(listener)

    def get_listeners_ids(self) -> list[str]:
        return [l.id for l in self.__connection_manager.get_listeners().values()]



import asyncio
import json
import signal

from asyncio import StreamReader, StreamWriter, start_unix_server
from asyncio.exceptions import CancelledError
from os import linesep
from pathlib import Path
from typing import AsyncIterable, Iterable

from eric_sse import get_logger
from eric_sse.entities import AbstractChannel
from eric_sse.message import MessageContract, Message
from eric_sse.exception import InvalidChannelException, InvalidMessageFormat
from eric_sse.prefabs import SSEChannel

logger = get_logger()


class ChannelContainer:
    """Helper class for management of multiple channels cases of use."""

    def __init__(self):
        self.__channels: dict[str: AbstractChannel] = {}


    def register(self, channel: AbstractChannel) -> None:
        if channel.id in self.__channels:
            raise InvalidChannelException(f'Channel with id {channel.id} already exists')
        self.__channels[channel.id] = channel

    def register_iterable(self, channels: Iterable[AbstractChannel]) -> None:
        registered_ids = set(self.get_all_ids())

        for channel in [c for c in channels if c.id not in registered_ids]:
            self.register(channel)

    def get(self, channel_id: str) -> AbstractChannel:
        try:
            return self.__channels[channel_id]
        except KeyError:
            raise InvalidChannelException(f'No channel with id {channel_id}')

    def rm(self, channel_id: str):
        try:
            del self.__channels[channel_id]
        except KeyError:
            raise InvalidChannelException(f'No channel with id {channel_id}')

    def get_all_ids(self) -> Iterable[str]:
        return self.__channels.keys()


class SocketServer:
    """
    An implementation of a socket server that acts as a controller to interact with library

    **Accepted format**: a plain JSON with the following keys::

        {        
            "c": "channel id" 
            "v": "verb" 
            "t": "message type" 
            "p": "message payload" 
            "r": "receiver (listener id when verb is 'rl')"
        }


    Possible values of **verb** identifies a supported action::

        "d" dispatch
        "b" broadcast
        "c" create channel
        "r" add listener
        "l" listen (opens a stream)
        "w" watch (opens a stream)
        "rl" remove a listener
        "rc" remove a channel

    See examples
    """
    cc = ChannelContainer()
    ACK = 'ack'

    def __init__(self, file_descriptor_path: str):
        """
        :param file_descriptor_path: See **start** method
        """
        self.__file_descriptor_path = file_descriptor_path
        self.__unix_server: asyncio.Server | None = None


    @staticmethod
    def start(file_descriptor_path: str):
        """
        Shortcut to start a server given a file descriptor path

        :param file_descriptor_path: file descriptor path, all understood by `Path <https://docs.python.org/3/library/pathlib.html#pathlib.Path>`_ is fine
        """
        logger.info('starting')
        try:
            server = SocketServer(file_descriptor_path)
            asyncio.run(server.main())
        except CancelledError:
            exit(0)

    @staticmethod
    def __parse(json_raw: str) -> (str, str, MessageContract | None, str):
        try:
            parsed = json.loads(json_raw)
            verb = parsed['v']

            channel_id = parsed.get('c')
            message = None
            if verb not in ['w', 'c', 'l', 'r', 'rl', 'rc']:
                message = Message(msg_type=parsed['t'], msg_payload=parsed.get('p'))

            receiver_id: str = parsed.get('r')
            return channel_id, verb, message, receiver_id

        except KeyError as e:
            logger.error(repr(e))
            raise InvalidMessageFormat(json_raw)

    @staticmethod
    async def connect_callback(reader: StreamReader, writer: StreamWriter):
        try:
            message_content = await reader.read()
            async for response in SocketServer.handle_command(message_content.decode()):
                writer.write(response.encode())
            writer.write_eof()
            await writer.drain()

        except Exception as e:
            logger.error(e)
            writer.write(repr(e).encode())
            writer.write_eof()
            await writer.drain()

    @staticmethod
    async def handle_command(raw_command: str) -> AsyncIterable[str]:
        channel_id, verb, message, receiver_id = SocketServer.__parse(raw_command)

        logger.info(f'received command {verb}')

        if verb == 'd':
            SocketServer.cc.get(channel_id).dispatch(receiver_id, message)
            yield SocketServer.ACK

        elif verb == 'c':
            channel = SSEChannel()
            SocketServer.cc.register(channel)
            yield channel.id

        elif verb == 'b':
            SocketServer.cc.get(channel_id).broadcast(message)
            yield SocketServer.ACK

        elif verb == 'r':
            l = SocketServer.cc.get(channel_id).add_listener()
            yield l.id

        elif verb == 'rl':
            SocketServer.cc.get(channel_id).remove_listener(listener_id=receiver_id)
            yield SocketServer.ACK

        elif verb == 'rc':
            SocketServer.cc.rm(channel_id)
            yield SocketServer.ACK

        elif verb == 'l':
            logger.info(f"Started listener {receiver_id} on {channel_id}")
            channel = SocketServer.cc.get(channel_id)
            listener = channel.get_listener(receiver_id)
            listener.start()
            async for m in channel.message_stream(listener):
                yield f'{json.dumps(m)}{linesep}'

        elif verb == 'w':
            logger.info(f"Client watching channel {channel_id}")
            channel = SocketServer.cc.get(channel_id)
            async for m in await channel.watch():
                yield f'{json.dumps(m)}{linesep}'

    async def shutdown(self):
        """Graceful Shutdown"""
        server = self.__unix_server
        logger.info("graceful shutdown")
        server.close()
        await server.wait_closed()
        Path(self.__file_descriptor_path).unlink()
        logger.info("done")

    async def main(self):
        """Starts the server"""
        server = await start_unix_server(SocketServer.connect_callback, path=Path(self.__file_descriptor_path))
        self.__unix_server = server
        addr = server.sockets[0].getsockname()
        logger.info(f'Serving on {addr}')

        async with server:
            for sig in (signal.SIGINT, signal.SIGTERM):
                server.get_loop().add_signal_handler(sig, lambda: asyncio.ensure_future(self.shutdown()))
            await server.serve_forever()

